                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*  version 7: iterative constructs  */
                     /*-----------------------------------*/
                     /*              evaluate             */
                     /*-----------------------------------*/

#include <stdio.h>

#include "SlipMain.h"

#include "SlipDictionary.h"
#include "SlipEvaluate.h"
#include "SlipThread.h"

/*------------------------------ private macros ------------------------------*/

#define chunk_size(TYP)                                                        \
  ((sizeof(TYP) - sizeof(CNT))/ sizeof(CEL_type))

/*------------------------------ private constants ---------------------------*/

static const TXT_type BRQ_error_string = "boolean value required";
static const TXT_type ILT_error_string = "illegal expression type";
static const TXT_type ITA_error_string = "improperly terminated application";
static const TXT_type IXT_error_string = "invalid expression type";
static const TXT_type PNR_error_string = "procedure or native required";
static const TXT_type TFO_error_string = "too few operands in application";
static const TXT_type TMO_error_string = "too many operands in application";

/*------------------------------ private prototypes --------------------------*/

static EXP_type        evaluate_bindings(PRC_type,
                                         VEC_type);
static EXP_type            evaluate_body(VEC_type,
                                         ENV_type);
static EXP_type      evaluate_expression(EXP_type);
static EXP_type     evaluate_native_call(NAT_type,
                                         VEC_type);
static EXP_type        evaluate_sequence(VEC_type);
static EXP_type evaluate_vararg_bindings(PRZ_type,
                                         VEC_type);

/*------------------------------ private functions ---------------------------*/

static EXP_type continue_with(EXP_type Value)
  { CNT_type continuation;
    CCC_type function;
    CFN_type c_function;
    continuation = Thread_Peek();
    c_function = continuation->cfn;
    function = c_function->ccc;
    return function(Value); }

static SGN_type list_length(PAI_type List)
  { PAI_type pair;
    SGN_type count;
    count = 0;
    for (pair = List;
         is_PAI(pair);
         pair = pair->cdr)
      count++;
    if (is_NUL(pair))
      return count;
    return -1; }

/*----------------------------------------------------------------------------*/
/*------------------------------------ application ---------------------------*/
/*----------------------------------------------------------------------------*/

static const TXT_type Application_String = "application";

static CFN_type Continue_application;

typedef struct aPL * aPL_type;
typedef struct aPL { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     VEC_type opd; } aPL;

static const UNS_type aPL_size = chunk_size(aPL);

static EXP_type continue_application(EXP_type Procedure)
  { aPL_type application_thread;
    TAG_type tag;
    VEC_type operands;
    application_thread = (aPL_type)Thread_Peek();
    operands = application_thread->opd;
    tag = Tag_of(Procedure);
    switch (tag)
      { case NAT_tag:
              return evaluate_native_call(Procedure,
                                          operands); 
        case PRC_tag:
                 return evaluate_bindings(Procedure,
                                          operands);  
        case PRZ_tag:
          return evaluate_vararg_bindings(Procedure,
                                          operands); }
    return Main_Error_Text(PNR_error_string,
                           Application_String); }
                           
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_application(APL_type Application)
  { aPL_type application_thread;
    EXP_type expression;
    VEC_type operands;
    expression = Application->exp;
    operands   = Application->opr;
    application_thread = (aPL_type)Thread_Push(Continue_application,
                                               aPL_size);
    application_thread->opd = operands;
    return evaluate_expression(expression); }

static NIL_type initialize_application(NIL_type)
  { Continue_application = make_CFN(continue_application); }
    
/*----------------------------------------------------------------------------*/
/*------------------------------------ begin ---------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_begin(BEG_type Begin)
  { VEC_type sequence;
    sequence = Begin->seq;
    return evaluate_sequence(sequence); }
      
/*----------------------------------------------------------------------------*/
/*---------------------------------- bindings --------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_bindings;

typedef struct bND * bND_type;
typedef struct bND { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     PRC_type prc;
                     VEC_type opd;
                     ENV_type clo;
                     NBR_type pos; } bND;

static const UNS_type bND_size = chunk_size(bND);

static EXP_type continue_bindings(EXP_type Value)
  { bND_type binding_thread;
    ENV_type closure;
    EXP_type expression;
    NBR_type position;
    PRC_type procedure;
    SYM_type parameter;
    VEC_type body,
             operands,
             parameters;
    UNS_type index,
             size_p;
    binding_thread = (bND_type)Thread_Peek();
    procedure = binding_thread->prc;
    closure   = binding_thread->clo;
    position  = binding_thread->pos;
    parameters = procedure->par;
    index = position->lng;
    parameter = parameters[index++];
    closure = Dictionary_Bind(parameter,
                              Value,
                              closure);
    size_p = size_VEC(parameters);
    if (index <= size_p)
      { binding_thread = (bND_type)Thread_Keep();
        position = make_NBR(index);
        operands = binding_thread->opd;
        binding_thread->clo = closure;
        binding_thread->pos = position;
        expression = operands[index];
        return evaluate_expression(expression); }
    body = procedure->bod;
    return evaluate_body(body,
                         closure); }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_bindings(PRC_type Procedure,
                                  VEC_type Operands)
  { bND_type binding_thread;
    ENV_type closure;
    EXP_type expression,
             parameters;
    UNS_type size_o,
             size_p;
    VEC_type body;
    parameters = Procedure->par;
    closure    = Procedure->clo;
    size_p = size_VEC(parameters);
    size_o = size_VEC(Operands);
    if (size_p > size_o)
      return Main_Error_Procedure(TFO_error_string,
                                  Procedure);
    if (size_p < size_o)
      return Main_Error_Procedure(TMO_error_string,
                                  Procedure);
    if (size_p == 0)
      { body = Procedure->bod;
        return evaluate_body(body,
                             closure); }
    binding_thread = (bND_type)Thread_Poke(Continue_bindings,
                                           bND_size);
    binding_thread->prc = Procedure;
    binding_thread->opd = Operands;
    binding_thread->clo = closure;
    binding_thread->pos = Main_One;
    expression = Operands[1];
    return evaluate_expression(expression); }

static NIL_type initialize_bindings(NIL_type)
  { Continue_bindings = make_CFN(continue_bindings); }

/*----------------------------------------------------------------------------*/
/*-------------------------------------- body --------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_body;

typedef struct bOD * bOD_type;
typedef struct bOD { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     ENV_type env; } bOD;

static const UNS_type bOD_size = chunk_size(bOD);

static EXP_type continue_body(EXP_type Value)
  { bOD_type body_thread;
    body_thread = (bOD_type)Thread_Pop();
    Dictionary_Environment = body_thread->env;
    return Value; }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_body(VEC_type Body,
                              ENV_type Closure)
  { bOD_type body_thread;
    body_thread = (bOD_type)Thread_Poke(Continue_body,
                                        bOD_size);
    body_thread->env = Dictionary_Environment;
    Dictionary_Environment = Closure;
    return evaluate_sequence(Body); }

static EXP_type evaluate_direct_body(VEC_type Body,
                                     ENV_type Closure)
  { bOD_type body_thread;
    body_thread = (bOD_type)Thread_Push(Continue_body,
                                        bOD_size);
    body_thread->env = Dictionary_Environment;
    Dictionary_Environment = Closure;
    return evaluate_sequence(Body); }

static EXP_type evaluate_inline_poke(EXP_type Expression)
  { bOD_type body_thread;
    body_thread = (bOD_type)Thread_Poke(Continue_body,
                                        bOD_size);
    body_thread->env = Dictionary_Environment;
    return evaluate_expression(Expression); }

static EXP_type evaluate_inline_push(EXP_type Expression)
  { bOD_type body_thread;
    body_thread = (bOD_type)Thread_Push(Continue_body,
                                        bOD_size);
    body_thread->env = Dictionary_Environment;
    return evaluate_expression(Expression); }

static EXP_type evaluate_inline_sequence(VEC_type Expressions)
  { bOD_type body_thread;
    body_thread = (bOD_type)Thread_Push(Continue_body,
                                        bOD_size);
    body_thread->env = Dictionary_Environment;
    return evaluate_sequence(Expressions); }

static NIL_type initialize_body(NIL_type)
  { Continue_body = make_CFN(continue_body); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ define --------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_define;

typedef struct dEF * dEF_type;
typedef struct dEF { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     SYM_type var; } dEF;

static const UNS_type dEF_size = chunk_size(dEF);

static EXP_type continue_define(EXP_type Value)
  { dEF_type define_thread;
    SYM_type variable;
    define_thread = (dEF_type)Thread_Pop();
    variable = define_thread->var;
    Dictionary_Define(variable,
                      Value);
    return Value; }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_define_function(DFF_type Define)
  { PRC_type procedure;
    SYM_type variable;
    VEC_type body,
             parameters;
    variable   = Define->var;
    parameters = Define->par;
    body       = Define->bod; 
    procedure = make_PRC(variable,
                         parameters,
                         body,
                         Main_Empty_Environment);
    Dictionary_Define(variable,
                      procedure);
    procedure->clo = Dictionary_Environment;
    return procedure; }
    
static EXP_type evaluate_define_function_vararg(DFZ_type Define)
  { PRZ_type procedure;
    SYM_type vararg,
             variable;
    VEC_type body,
             parameters;
    variable   = Define->var;
    parameters = Define->par;
    vararg     = Define->vrg;
    body       = Define->bod; 
    procedure = make_PRZ(variable,
                         parameters,
                         vararg,
                         body,
                         Main_Empty_Environment);
    Dictionary_Define(variable,
                      procedure);
    procedure->clo = Dictionary_Environment;
    return procedure; }
        
static EXP_type evaluate_define_variable(DFV_type Define)
  { dEF_type define_thread;
    EXP_type expression;
    SYM_type variable;
    variable   = Define->var;
    expression = Define->exp;
    define_thread = (dEF_type)Thread_Push(Continue_define,
                                          dEF_size);
    define_thread->var = variable;
    return evaluate_expression(expression); }
    
static NIL_type initialize_define(NIL_type)
  { Continue_define = make_CFN(continue_define); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ direct_application --------------------*/
/*----------------------------------------------------------------------------*/

static const TXT_type Direct_Application_String = "direct application";

static EXP_type evaluate_direct_application(EXP_type Value,
                                            PAI_type Arguments)
  { EXP_type value;
    ENV_type closure;
    FUN_type function;
    NAT_type native;
    PAI_type arguments;
    PRC_type procedure;
    PRZ_type procedure_vararg;
    SGN_type length;
    SYM_type variable,
             vararg;
    TAG_type tag;
    TXT_type name;
    UNS_type index,
             size_p;
    VEC_type body,
             parameters,
             vector;
    tag = Tag_of(Value);
    switch (tag)
      { case NAT_tag:
          native = (NAT_type)Value;
          function = native->fun;
          length = list_length(Arguments);
          switch (length)
            { case 0:
                vector = Main_Empty_Vector;
                break;
              case -1:
                { name = native->nam;
                  return Main_Error_Text(ITA_error_string,
                                         name); }
              default:
                { arguments = Arguments;
                  vector = make_VEC(length);
                  for (index = 1;
                       index <= length;
                       index++)
                    { value = arguments->car;
                      arguments = arguments->cdr;
                      vector[index] = value; }
                  break; }}
          return function(vector); 
        case PRC_tag:
          arguments = Arguments;
          procedure = (PRC_type)Value;
          parameters = procedure->par;
          closure    = procedure->clo;
          size_p = size_VEC(parameters);
          for (index = 1; 
               index <= size_p;
               index++)
            { variable = parameters[index];
              if (is_NUL(arguments))
                return Main_Error_Procedure(TFO_error_string, 
                                            procedure);
              value = arguments->car;
              arguments = arguments->cdr;
              closure = Dictionary_Bind(variable,
                                        value,
                                        closure); }
          if (!is_NUL(arguments))
            return Main_Error_Procedure(TMO_error_string,
                                        procedure);
          body = procedure->bod;
          return evaluate_direct_body(body,
                                      closure); 
        case PRZ_tag:
          arguments = Arguments;
          procedure_vararg = (PRZ_type)Value;
          parameters = procedure_vararg->par;
          vararg     = procedure_vararg->vrg;
          closure    = procedure_vararg->clo;
          size_p = size_VEC(parameters);
          for (index = 1; 
               index <= size_p;
               index++)
            { variable = parameters[index];
              if (is_NUL(arguments))
                return Main_Error_Procedure_Vararg(TFO_error_string, 
                                                   procedure_vararg);
              value = arguments->car;
              arguments = arguments->cdr;
              closure = Dictionary_Bind(variable,
                                        value,
                                        closure); }
          closure = Dictionary_Bind(vararg,
                                    arguments,
                                    closure); 
          body = procedure_vararg->bod;
          return evaluate_direct_body(body,
                                      closure); }
    return Main_Error_Text(PNR_error_string,
                           Direct_Application_String); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ if ------------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_if_double;

typedef struct iFD * iFD_type;
typedef struct iFD { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     EXP_type cns;
                     EXP_type alt; } iFD;

static const UNS_type iFD_size = chunk_size(iFD);

static EXP_type continue_if_double(EXP_type Boolean)
  { iFD_type if_thread;
    EXP_type alternative,
             consequent;
    TAG_type tag;
    tag = Tag_of(Boolean);
    if_thread = (iFD_type)Thread_Peek();
    switch (tag)
      { case FLS_tag:
          alternative = if_thread->alt;
          return evaluate_inline_poke(alternative);
        case TRU_tag:
          consequent = if_thread->cns;
          return evaluate_inline_poke(consequent); }
    return Main_Error_Text(BRQ_error_string,
                           Main_If_String); }

/*----------------------------------------------------------------------------*/

static CFN_type Continue_if_single;

typedef struct iFS * iFS_type;
typedef struct iFS { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     EXP_type cns; } iFS;

static const UNS_type iFS_size = chunk_size(iFS);

static EXP_type continue_if_single(EXP_type Boolean)
  { iFS_type if_single_thread;
    EXP_type consequent;
    TAG_type tag;
    tag = Tag_of(Boolean);
    if_single_thread = (iFS_type)Thread_Peek();
    switch (tag)
      { case FLS_tag:
          Thread_Zap();
          return Main_Unspecified;
        case TRU_tag:
          consequent = if_single_thread->cns;
          return evaluate_inline_poke(consequent); }
    return Main_Error_Text(BRQ_error_string,
                           Main_If_String); }
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_if_double(IFF_type If)
  { iFD_type if_thread;
    EXP_type alternative,
             consequent,
             predicate;
    predicate   = If->prd;
    consequent  = If->cns;
    alternative = If->alt;
    if_thread = (iFD_type)Thread_Push(Continue_if_double,
                                      iFD_size);
    if_thread->cns = consequent;
    if_thread->alt = alternative;
    return evaluate_expression(predicate); }
    
static EXP_type evaluate_if_single(IFZ_type If)
  { iFS_type if_thread;
    EXP_type consequent,
             predicate;
    predicate  = If->prd;
    consequent = If->cns;
    if_thread = (iFS_type)Thread_Push(Continue_if_single,
                                      iFS_size);
    if_thread->cns = consequent;
    return evaluate_expression(predicate); }
    
static NIL_type initialize_if(NIL_type)
  { Continue_if_double = make_CFN(continue_if_double);
    Continue_if_single = make_CFN(continue_if_single); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ lambda --------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_lambda(LMB_type Lambda)
  { PRC_type procedure;
    VEC_type parameters;
    VEC_type body;
    parameters = Lambda->par;
    body       = Lambda->bod;
    procedure = make_PRC(Main_Anonymous,
                         parameters,
                         body,
                         Dictionary_Environment);
    return procedure; }
    
static EXP_type evaluate_lambda_vararg(LMZ_type Lambda)
  { PRZ_type procedure;
    SYM_type vararg;
    VEC_type parameters;
    VEC_type body;
    parameters = Lambda->par;
    vararg     = Lambda->vrg;
    body       = Lambda->bod;
    procedure = make_PRZ(Main_Anonymous,
                         parameters,
                         vararg,
                         body,
                         Dictionary_Environment);
    return procedure; }

/*----------------------------------------------------------------------------*/
/*-------------------------------- native call -------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_native_call;

typedef struct nCL * nCL_type;
typedef struct nCL { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     VEC_type opd;
                     VEC_type vec;
                     NAT_type nat;
                     NBR_type pos; } nCL;

static const UNS_type nCL_size = chunk_size(nCL);

static EXP_type continue_native_call(EXP_type Value)
  { nCL_type native_call_thread;
    EXP_type expression;
    FUN_type function;
    NAT_type native;
    NBR_type position;
    UNS_type index,
             size_v;
    VEC_type operands,
             vector;
    native_call_thread = (nCL_type)Thread_Peek();
    vector   = native_call_thread->vec;
    position = native_call_thread->pos;
    index = position->lng;
    size_v = size_VEC(vector);
    vector[index] = Value;
    if (index++ < size_v)
      { native_call_thread = (nCL_type)Thread_Keep();
        operands = native_call_thread->opd;
        expression = operands[index];
        position = make_NBR(index);
        native_call_thread->pos = position;
        return evaluate_expression(expression); }
    native = native_call_thread->nat;
    function = native->fun;
    Thread_Zap();
    return function(vector); }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_native_call(NAT_type Native,
                                     VEC_type Operands)
  { nCL_type native_call_thread;
    EXP_type expression;
    FUN_type function;
    UNS_type size_o;
    VEC_type vector;
    size_o = size_VEC(Operands);
    if (size_o == 0)
      { function = Native->fun;
        Thread_Zap();
        return function(Main_Empty_Vector); }
    vector = make_VEC(size_o);
    native_call_thread = (nCL_type)Thread_Poke(Continue_native_call,
                                               nCL_size);
    native_call_thread->opd = Operands;
    native_call_thread->vec = vector;
    native_call_thread->nat = Native;
    native_call_thread->pos = Main_One;
    expression = Operands[1];
    return evaluate_expression(expression); }
                       
static NIL_type initialize_native_call(NIL_type)
  { Continue_native_call = make_CFN(continue_native_call); }
  
/*----------------------------------------------------------------------------*/
/*------------------------------------ quote ---------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_quote(QUO_type Quote)
  { EXP_type expression;
    expression = Quote->exp;
    return expression; }        /* clone!!! */

/*----------------------------------------------------------------------------*/
/*---------------------------------- sequence --------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_sequence;

typedef struct sEQ * sEQ_type;
typedef struct sEQ { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     VEC_type exs;
                     NBR_type pos; } sEQ;

static const UNS_type sEQ_size = chunk_size(sEQ);

static EXP_type continue_sequence(EXP_type Value)
  { sEQ_type sequence_thread;
    EXP_type expression;
    VEC_type expressions;
    NBR_type position;
    UNS_type index,
             size_x;
    sequence_thread = (sEQ_type)Thread_Peek();
    expressions = sequence_thread->exs;
    position    = sequence_thread->pos;
    index = position->lng + 1;
    expression = expressions[index];
    size_x = size_VEC(expressions);
    if (index < size_x)
      { sequence_thread = (sEQ_type)Thread_Keep();
        position = make_NBR(index);
        sequence_thread->pos = position; }
    else
      Thread_Zap();
    return evaluate_expression(expression); }
    
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_sequence(VEC_type Expressions)
  { sEQ_type sequence_thread;
    EXP_type expression;
    UNS_type size_x;
    size_x = size_VEC(Expressions);
    expression = Expressions[1];
    if (size_x > 1)
      { sequence_thread = (sEQ_type)Thread_Push(Continue_sequence,
                                                sEQ_size);
        sequence_thread->exs = Expressions;
        sequence_thread->pos = Main_One; }
    return evaluate_expression(expression); }
    
static NIL_type initialize_sequence(NIL_type)
  { Continue_sequence = make_CFN(continue_sequence); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ set -----------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_set;

typedef struct sET * sET_type;
typedef struct sET { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     SYM_type var; } sET;

static const UNS_type sET_size = chunk_size(sET);

static EXP_type continue_set(EXP_type Value)
  { sET_type set_thread;
    SYM_type variable;
    set_thread = (sET_type)Thread_Pop();
    variable = set_thread->var;
    Dictionary_Replace(variable,
                       Value);
    return Value; }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_set(SET_type Set)
  { sET_type set_thread;
    EXP_type expression;
    SYM_type variable;
    variable   = Set->var;
    expression = Set->exp;
    set_thread = (sET_type)Thread_Push(Continue_set,
                                       sET_size);
    set_thread->var = variable;
    return evaluate_expression(expression); }
    
static NIL_type initialize_set(NIL_type)
  { Continue_set = make_CFN(continue_set); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ symbol --------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_symbol(SYM_type Variable)
  { EXP_type value;
    value = Dictionary_Lookup(Variable);
    return value; }

/*----------------------------------------------------------------------------*/
/*------------------------------------ value ---------------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_value(EXP_type Value)
  { return Value; }

/*----------------------------------------------------------------------------*/
/*------------------------------- vararg_bindings ----------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_vararg,
                Continue_vararg_bindings;

typedef struct vRG * vRG_type;
typedef struct vRG { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     PRZ_type prc;
                     VEC_type opd;
                     PAI_type lst;
                     ENV_type clo;
                     NBR_type pos; } vRG;

static const UNS_type vRG_size = chunk_size(vRG);

static EXP_type continue_vararg(EXP_type Value)
  { vRG_type binding_thread;
    ENV_type closure;
    EXP_type expression;
    NBR_type position;
    PAI_type list;
    PRZ_type procedure;
    SYM_type vararg;
    VEC_type operands;
    UNS_type index,
             size_o;
    VEC_type body;
    binding_thread = (vRG_type)Thread_Peek();
    operands  = binding_thread->opd;
    list      = binding_thread->lst;
    position  = binding_thread->pos;
    index = position->lng;
    size_o = size_VEC(operands);
    list = make_PAI(Value,
                    list);                             
    if (index < size_o)
      { binding_thread = (vRG_type)Thread_Keep();
        position = make_NBR(++index);
        binding_thread->lst = list;
        binding_thread->pos = position;
        expression = operands[index];
        return evaluate_expression(expression); }
    procedure = binding_thread->prc;
    closure   = binding_thread->clo;
    vararg = procedure->vrg;
    body   = procedure->bod;
    list = Main_Reverse(list);
    closure = Dictionary_Bind(vararg,
                              list,
                              closure);
    return evaluate_body(body,
                         closure); }

static EXP_type continue_vararg_bindings(EXP_type Value)
  { vRG_type binding_thread;
    ENV_type closure;
    EXP_type expression;
    NBR_type position;
    PRZ_type procedure;
    SYM_type parameter,
             vararg;
    VEC_type operands,
             parameters;
    UNS_type index,
             size_o,
             size_p;
    VEC_type body;
    binding_thread = (vRG_type)Thread_Peek();
    procedure = binding_thread->prc;
    operands  = binding_thread->opd;
    closure   = binding_thread->clo;
    position  = binding_thread->pos;
    parameters = procedure->par;
    index = position->lng;
    parameter = parameters[index];
    closure = Dictionary_Bind(parameter,
                              Value,
                              closure); 
    size_o = size_VEC(operands);
    if (index == size_o)
      { vararg = procedure->vrg;
        body   = procedure->bod;
        closure = Dictionary_Bind(vararg,
                                  Main_Empty_Pair,
                                  closure);
        return evaluate_body(body,
                             closure); }
    size_p = size_VEC(parameters);
    if (index == size_p)
      binding_thread = (vRG_type)Thread_Patch(Continue_vararg);
    position = make_NBR(++index);
    binding_thread->clo = closure;
    binding_thread->pos = position;
    expression = operands[index];
    return evaluate_expression(expression); }
    
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_vararg_bindings(PRZ_type Procedure,
                                         VEC_type Operands)
  { vRG_type binding_thread;
    CFN_type c_function;
    ENV_type closure;
    EXP_type expression,
             parameters;
    SYM_type vararg;
    UNS_type size_p,
             size_o;
    VEC_type body;
    parameters = Procedure->par;
    size_p = size_VEC(parameters);
    size_o = size_VEC(Operands);
    if (size_p > size_o)
      return Main_Error_Procedure_Vararg(TFO_error_string,
                                         Procedure);
    closure = Procedure->clo;
    if (size_o == 0)
      { vararg = Procedure->vrg;
        body   = Procedure->bod;
        closure = Dictionary_Bind(vararg,
                                  Main_Empty_Pair,
                                  closure);
        return evaluate_body(body,
                             closure); }
    if (size_p == 0)
      c_function = Continue_vararg;
    else
      c_function = Continue_vararg_bindings;
    binding_thread = (vRG_type)Thread_Poke(c_function,
                                           vRG_size);
    binding_thread->prc = Procedure;
    binding_thread->opd = Operands;
    binding_thread->lst = Main_Empty_Pair;
    binding_thread->clo = closure;
    binding_thread->pos = Main_One;
    expression = Operands[1];
    return evaluate_expression(expression); }
                
static NIL_type initialize_vararg_bindings(NIL_type)
  { Continue_vararg = make_CFN(continue_vararg);
    Continue_vararg_bindings = make_CFN(continue_vararg_bindings); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ while ---------------------------------*/
/*----------------------------------------------------------------------------*/

static CFN_type Continue_while_body,
                Continue_while_predicate;

typedef struct wHI * wHI_type;
typedef struct wHI { CEL_type hdr;
                     CFN_type cfn;
                     CNT_type cnt;
                     EXP_type prd;
                     VEC_type bod;
                     ENV_type env;
                     EXP_type res; } wHI;

static const UNS_type wHI_size = chunk_size(wHI);

static EXP_type continue_while_predicate(EXP_type Boolean)
  { wHI_type while_thread;
    ENV_type environment;
    EXP_type result;
    VEC_type body;
    while_thread = (wHI_type)Thread_Peek();
    environment = while_thread->env;
    Dictionary_Environment = environment;
    if (is_FLS(Boolean))
      { result = while_thread->res;
        Thread_Zap();
        return result; }
    while_thread = (wHI_type)Thread_Patch(Continue_while_body);
    body = while_thread->bod;
    return evaluate_inline_sequence(body); }

static EXP_type continue_while_body(EXP_type Value)
  { wHI_type while_thread;
    ENV_type environment;
    EXP_type predicate;
    while_thread = (wHI_type)Thread_Peek();
    predicate   = while_thread->prd;
    environment = while_thread->env;
    Dictionary_Environment = environment;
    while_thread = (wHI_type)Thread_Patch(Continue_while_predicate);
    while_thread->res = Value;
    return evaluate_inline_push(predicate); }

/*----------------------------------------------------------------------------*/

static EXP_type evaluate_while(WHI_type While)
  { wHI_type while_thread;
    EXP_type predicate;
    VEC_type body;
    predicate = While->prd;
    body      = While->bod;
    while_thread = (wHI_type)Thread_Push(Continue_while_predicate,
                                         wHI_size);
    while_thread->prd = predicate;
    while_thread->bod = body;
    while_thread->env = Dictionary_Environment;
    while_thread->res = Main_Unspecified;
    return evaluate_inline_push(predicate); }
                                             
static NIL_type initialize_while(NIL_type)
  { Continue_while_predicate = make_CFN(continue_while_predicate);
    Continue_while_body = make_CFN(continue_while_body); }

/*----------------------------------------------------------------------------*/
/*------------------------------------ expression ----------------------------*/
/*----------------------------------------------------------------------------*/

static EXP_type evaluate_expression(EXP_type Expression)
  { TAG_type tag;
    tag = Tag_of(Expression);
    switch (tag)
      { case APL_tag:
                     return evaluate_application(Expression);
        case BEG_tag:
                           return evaluate_begin(Expression);
        case DFF_tag:
                 return evaluate_define_function(Expression);
        case DFV_tag:
                 return evaluate_define_variable(Expression);
        case DFZ_tag:
          return evaluate_define_function_vararg(Expression);
        case IFF_tag:
                       return evaluate_if_double(Expression);
        case IFZ_tag:
                       return evaluate_if_single(Expression);
        case LMB_tag:
                          return evaluate_lambda(Expression);
        case LMZ_tag:
                   return evaluate_lambda_vararg(Expression);
        case QUO_tag:
                           return evaluate_quote(Expression);
        case SET_tag:
                             return evaluate_set(Expression);
        case SYM_tag:
                          return evaluate_symbol(Expression);
        case WHI_tag:
                           return evaluate_while(Expression);
        case CHA_tag:
        case FLS_tag:
        case NAT_tag:
        case NBR_tag:
        case NUL_tag:
        case PAI_tag:
        case PRC_tag:
        case PRZ_tag:
        case REA_tag:
        case STR_tag:
        case TRU_tag:
        case USP_tag:
        case VEC_tag:
                           return evaluate_value(Expression);
        case CFN_tag:
        case CNT_tag:
        case ENV_tag:
          return Main_Error_Tag(ILT_error_string, 
                                tag); }
    return Main_Error_Tag(IXT_error_string,
                          tag); }

/*------------------------------ public functions ----------------------------*/

EXP_type Evaluate_Apply(EXP_type Procedure,
                        PAI_type Arguments)
  { return evaluate_direct_application(Procedure,
                                       Arguments); }

EXP_type Evaluate_Continue(EXP_type Value)
  { return continue_with(Value); }

EXP_type Evaluate_Evaluate(EXP_type Expression)
  { return evaluate_expression(Expression); }

NIL_type Evaluate_Initialize(NIL_type)
  { initialize_application();
    initialize_bindings();
    initialize_body();
    initialize_define();
    initialize_if();
    initialize_native_call();
    initialize_sequence();
    initialize_set();
    initialize_vararg_bindings();
    initialize_while(); }
