                     /*-----------------------------------*/
                     /*             >>>Slip<<<            */
                     /*            Theo D'Hondt           */
                     /*     VUB Software Languages Lab    */
                     /*             (c) 2010              */
                     /*-----------------------------------*/
                     /*  version 7: iterative constructs  */
                     /*-----------------------------------*/
                     /*               grammar             */
                     /*-----------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "SlipGrammar.h"

/*------------------------------ private macros ------------------------------*/

#define chunk_size(TYP)                                                        \
  (sizeof(TYP) / sizeof(CEL_type) - 1)

#define make_chunk_with_offset(TYP, SIZ)                                       \
  (TYP##_type)Memory_Make_Chunk(TYP##_tag, TYP##_size + SIZ)

#define make_chunk(TYP)                                                        \
  make_chunk_with_offset(TYP, 0)

/*------------------------------ private functions ---------------------------*/

static BYT_type member_of(EXP_type Expression,
                          TAG_type Tag)
  { BYT_type tag;
    tag = Tag_of(Expression);
    return (tag == Tag); }

static UNS_type string_to_expression_size(TXT_type Raw_string)
  { return strlen(Raw_string) / sizeof(CEL_type) + 1; }

/*-------------------------------------Tag -----------------------------------*/

TAG_type Tag_of(EXP_type Expression)
  { return (TAG_type)Memory_Get_Tag((PTR_type)Expression); }

/*------------------------------ APL declarations ----------------------------*/

static const UNS_type APL_size = chunk_size(APL);

BYT_type is_APL(EXP_type Expression)
  { BYT_type tag = Tag_of(Expression);
    return (tag == APL_tag); }

APL_type make_APL(EXP_type Expression,
                  VEC_type Operands)
  { APL_type application = make_chunk(APL);
    application->exp = Expression;
    application->opr = Operands;
    return application; }

/*------------------------------ BEG declarations ----------------------------*/

static const UNS_type BEG_size = chunk_size(BEG);

BYT_type is_BEG(EXP_type Expression)
  { BYT_type tag = Tag_of(Expression);
    return (tag == BEG_tag); }

BEG_type make_BEG(VEC_type Sequence)
  { BEG_type begin = make_chunk(BEG);
    begin->seq = Sequence;
    return begin; }

/*------------------------------ CFN declarations ----------------------------*/

static const UNS_type CFN_size = chunk_size(CFN);

BYT_type is_CFN(EXP_type Expression)
  { return member_of(Expression,
                     CFN_tag); }

CFN_type make_CFN(CCC_type C_function)
  { CFN_type c_function;
    c_function = make_chunk(CFN);
    c_function->ccc = C_function;
    return c_function; }

/*------------------------------ CHA declarations ----------------------------*/

static const UNS_type CHA_size = chunk_size(CHA);

BYT_type is_CHA(EXP_type Expression)
  { return member_of(Expression,
                     CHA_tag); }

CHA_type make_CHA(CHR_type Raw_character)
  { CHA_type character;
    character = make_chunk(CHA);
    character->chr = Raw_character;
    return character; }

/*------------------------------ CNT declarations ----------------------------*/

static const UNS_type CNT_size = chunk_size(CNT);

BYT_type is_CNT(EXP_type Expression)
  { return member_of(Expression,
                     CNT_tag); }

CNT_type make_CNT(CFN_type C_function,
                  CNT_type Continuation,
                  UNS_type Size)
  { CNT_type continuation;
    continuation = make_chunk_with_offset(CNT,
                                          Size);
    continuation->cfn = C_function;
    continuation->cnt = Continuation;
    return continuation; }

/*------------------------------ DFF declarations ----------------------------*/

static const UNS_type DFF_size = chunk_size(DFF);

BYT_type is_DFF(EXP_type Expression)
  { BYT_type tag = Tag_of(Expression);
    return (tag == DFF_tag); }

DFF_type make_DFF(SYM_type Variable,
                  VEC_type Parameters,
                  VEC_type Body)
  { DFF_type definition = make_chunk(DFF);
    definition->var = Variable;
    definition->par = Parameters;
    definition->bod = Body;
    return definition; }

/*------------------------------ DFV declarations ----------------------------*/

static const UNS_type DFV_size = chunk_size(DFV);

BYT_type is_DFV(EXP_type Expression)
  { BYT_type tag = Tag_of(Expression);
    return (tag == DFV_tag); }

DFV_type make_DFV(SYM_type Variable,
                  EXP_type Expression)
  { DFV_type definition = make_chunk(DFV);
    definition->var = Variable;
    definition->exp = Expression;
    return definition; }

/*------------------------------ DFZ declarations ----------------------------*/

static const UNS_type DFZ_size = chunk_size(DFZ);

BYT_type is_DFZ(EXP_type Expression)
  { BYT_type tag = Tag_of(Expression);
    return (tag == DFZ_tag); }

DFZ_type make_DFZ(SYM_type Variable,
                  VEC_type Parameters,
                  SYM_type Vararg,
                  VEC_type Body)
  { DFZ_type definition = make_chunk(DFZ);
    definition->var = Variable;
    definition->par = Parameters;
    definition->vrg = Vararg;
    definition->bod = Body;
    return definition; }

/*------------------------------ ENV declarations ----------------------------*/

static const UNS_type ENV_size = chunk_size(ENV);

BYT_type is_ENV(EXP_type Expression)
  { return member_of(Expression,
                     ENV_tag); }

ENV_type make_ENV(SYM_type Variable,
                  EXP_type Value,
                  ENV_type Environment)
  { ENV_type environment;
    environment = make_chunk(ENV);
    environment->var = Variable;
    environment->val = Value;
    environment->env = Environment;
    return environment; }

/*------------------------------ FLS declarations ----------------------------*/

static const UNS_type FLS_size = chunk_size(FLS);

BYT_type is_FLS(EXP_type Expression)
  { return member_of(Expression,
                     FLS_tag); }

FLS_type make_FLS(NIL_type)
  { FLS_type false;
    false = make_chunk(FLS);
    return false; }

/*------------------------------ IFF declarations ----------------------------*/

static const UNS_type IFF_size = chunk_size(IFF);

BYT_type is_IFF(EXP_type Expression)
  { BYT_type tag = Tag_of(Expression);
    return (tag == IFF_tag); }

IFF_type make_IFF(EXP_type Predicate,
                  EXP_type Consequent,
                  EXP_type Alternative)
  { IFF_type if_ = make_chunk(IFF);
    if_->prd = Predicate;
    if_->cns = Consequent;
    if_->alt = Alternative;
    return if_; }

/*------------------------------ IFZ declarations ----------------------------*/

static const UNS_type IFZ_size = chunk_size(IFZ);

BYT_type is_IFZ(EXP_type Expression)
  { BYT_type tag = Tag_of(Expression);
    return (tag == IFZ_tag); }

IFZ_type make_IFZ(EXP_type Predicate,
                  EXP_type Consequent)
  { IFZ_type if_ = make_chunk(IFZ);
    if_->prd = Predicate;
    if_->cns = Consequent;
    return if_; }

/*------------------------------ LMB declarations ----------------------------*/

static const UNS_type LMB_size = chunk_size(LMB);

BYT_type is_LMB(EXP_type Expression)
  { BYT_type tag = Tag_of(Expression);
    return (tag == LMB_tag); }

LMB_type make_LMB(VEC_type Parameters,
                  VEC_type Body)
  { LMB_type lambda = make_chunk(LMB);
    lambda->par = Parameters;
    lambda->bod = Body;
    return lambda; }

/*------------------------------ LMZ declarations ----------------------------*/

static const UNS_type LMZ_size = chunk_size(LMZ);

BYT_type is_LMZ(EXP_type Expression)
  { BYT_type tag = Tag_of(Expression);
    return (tag == LMZ_tag); }

LMZ_type make_LMZ(VEC_type Parameters,
                  SYM_type Vararg,
                  VEC_type Body)
  { LMZ_type lambda = make_chunk(LMZ);
    lambda->par = Parameters;
    lambda->vrg = Vararg;
    lambda->bod = Body;
    return lambda; }

/*------------------------------ NAT declarations ----------------------------*/

static const UNS_type NAT_size = chunk_size(NAT);

BYT_type is_NAT(EXP_type Expression)
  { return member_of(Expression,
                     NAT_tag); }

NAT_type make_NAT(FUN_type Raw_function,
                  TXT_type Name)
  { NAT_type native;
    UNS_type size;
    size = string_to_expression_size(Name);
    native = make_chunk_with_offset(NAT,
                                    size);
    native->fun = Raw_function;
    strcpy(native->nam,
           Name);
    return native; }

/*------------------------------ NBR declarations ----------------------------*/

static const UNS_type NBR_size = chunk_size(NBR);

BYT_type is_NBR(EXP_type Expression)
  { return member_of(Expression,
                     NBR_tag); }

NBR_type make_NBR(LNG_type Raw_number)
  { NBR_type number;
    number = make_chunk(NBR);
    number->lng = Raw_number;
    return number; }

/*------------------------------ NUL declarations ----------------------------*/

static const UNS_type NUL_size = chunk_size(NUL);

BYT_type is_NUL(EXP_type Expression)
   { return member_of(Expression,
                     NUL_tag); }

NUL_type make_NUL(NIL_type)
  { NUL_type null;
    null = make_chunk(NUL);
    return null; }

/*------------------------------ PAI declarations ----------------------------*/

static const UNS_type PAI_size = chunk_size(PAI);

BYT_type is_PAI(EXP_type Expression)
  { return member_of(Expression,
                     PAI_tag); }

PAI_type make_PAI(EXP_type Car,
                  EXP_type Cdr)
  { PAI_type pair;
    pair = make_chunk(PAI);
    pair->car = Car;
    pair->cdr = Cdr;
    return pair; }

/*------------------------------ PRC declarations ----------------------------*/

static const UNS_type PRC_size = chunk_size(PRC);

BYT_type is_PRC(EXP_type Expression)
  { return member_of(Expression,
                     PRC_tag); }

PRC_type make_PRC(SYM_type Name,
                  VEC_type Parameters,
                  VEC_type Body,
                  ENV_type Closure)
  { PRC_type procedure;
    procedure = make_chunk(PRC);
    procedure->nam = Name;
    procedure->par = Parameters;
    procedure->bod = Body;
    procedure->clo = Closure;
    return procedure; }

/*------------------------------ PRZ declarations ----------------------------*/

static const UNS_type PRZ_size = chunk_size(PRZ);

BYT_type is_PRZ(EXP_type Expression)
  { return member_of(Expression,
                     PRZ_tag); }

PRZ_type make_PRZ(SYM_type Name,
                  VEC_type Parameters,
                  SYM_type Vararg,
                  VEC_type Body,
                  ENV_type Closure)
  { PRZ_type procedure;
    procedure = make_chunk(PRZ);
    procedure->nam = Name;
    procedure->par = Parameters;
    procedure->vrg = Vararg;
    procedure->bod = Body;
    procedure->clo = Closure;
    return procedure; }

/*------------------------------ QUO declarations ----------------------------*/

static const UNS_type QUO_size = chunk_size(QUO);

BYT_type is_QUO(EXP_type Expression)
  { BYT_type tag = Tag_of(Expression);
    return (tag == QUO_tag); }

QUO_type make_QUO(EXP_type Expression)
  { QUO_type quote = make_chunk(QUO);
    quote->exp = Expression;
    return quote; }

/*------------------------------ REA declarations ----------------------------*/

static const UNS_type REA_size = chunk_size(REA);

BYT_type is_REA(EXP_type Expression)
  { return member_of(Expression,
                     REA_tag); }

REA_type make_REA(FLO_type Raw_real)
  { REA_type real;
    real = make_chunk(REA);
    real->flo = Raw_real;
    return real; }

/*------------------------------ SET declarations ----------------------------*/

static const UNS_type SET_size = chunk_size(SET);

BYT_type is_SET(EXP_type Expression)
  { BYT_type tag = Tag_of(Expression);
    return (tag == SET_tag); }

SET_type make_SET(SYM_type Variable,
                  EXP_type Expression)
  { SET_type set = make_chunk(SET);
    set->var = Variable;
    set->exp = Expression;
    return set; }

/*------------------------------ STR declarations ----------------------------*/

static const UNS_type STR_size = chunk_size(STR);

BYT_type is_STR(EXP_type Expression)
  { return member_of(Expression,
                     STR_tag); }

STR_type make_STR(TXT_type Raw_string)
  { STR_type string;
    UNS_type size;
    size = string_to_expression_size(Raw_string);
    string = make_chunk_with_offset(STR,
                                    size);
    strcpy(string->txt,
           Raw_string);
    return string; }

/*------------------------------ SYM declarations ----------------------------*/

static const UNS_type SYM_size = chunk_size(SYM);

BYT_type is_SYM(EXP_type Expression)
  { return member_of(Expression,
                     SYM_tag); }

SYM_type make_SYM(TXT_type Raw_string)
  { SYM_type symbol;
    UNS_type size;
    size = string_to_expression_size(Raw_string);
    symbol = make_chunk_with_offset(SYM,
                                    size);
    strcpy(symbol->txt,
           Raw_string);
    return symbol; }

/*------------------------------ TRU declarations ----------------------------*/

static const UNS_type TRU_size = chunk_size(TRU);

BYT_type is_TRU(EXP_type Expression)
  { return member_of(Expression,
                     TRU_tag); }

TRU_type make_TRU(NIL_type)
  { TRU_type true;
    true = make_chunk(TRU);
    return true; }

/*------------------------------ USP declarations ----------------------------*/

static const UNS_type USP_size = chunk_size(USP);

BYT_type is_USP(EXP_type Expression)
  { return member_of(Expression,
                     USP_tag); }

USP_type make_USP(NIL_type)
  { USP_type unspecified;
    unspecified = make_chunk(USP);
    return unspecified; }

/*------------------------------ VEC declarations ----------------------------*/

static const UNS_type VEC_size = 0;

BYT_type is_VEC(EXP_type Expression)
  { return member_of(Expression,
                     VEC_tag); }

VEC_type make_VEC(UNS_type Size)
  { VEC_type vector;
    vector = (VEC_type)make_chunk_with_offset(VEC,
                                              Size);
    return vector; }

UNS_type size_VEC(VEC_type Vector)
  { return (UNS_type) Memory_Get_Size((PTR_type)Vector); }

/*------------------------------ WHI declarations ----------------------------*/

static const UNS_type WHI_size = chunk_size(WHI);

BYT_type is_WHI(EXP_type Expression)
  { BYT_type tag = Tag_of(Expression);
    return (tag == WHI_tag); }

WHI_type make_WHI(EXP_type Predicate,
                  VEC_type Body)
  { WHI_type while_ = make_chunk(WHI);
    while_->prd = Predicate;
    while_->bod = Body;
    return while_; }
